/*	engine_position_isrs.c

	Copyright 2008 Fred Cooke

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

#include "engine_position_isrs.h"

/* TODO Summary of intended engine position capture scheme
 * 
 * Position/RPM signal interpretation :
 * Discard edges that have arrived too soon (lose sync here?)
 * Check to ensure we haven't lost sync (pulse arrives too late)
 * Compare time stamps of successive edges and calculate RPM
 * Store RPM and position in globals
 * 
 * Schedule events :
 * loop through all events (spark and fuel), schedule those that fall sufficiently after this tooth and before the next one we expect.
 * 
 * Sample ADCs :
 * Grab a unified set of ADC readings at one time in a consistent crank location to eliminate engine cycle dependent noise.
 * Set flag stating that New pulse, advance, etc should be calculated.
 * 
 */

void PrimaryRPMISR(void)
{
	/* Clear the interrupt flag for this input compare channel */
	TFLG = 0x01;

	/* Save all relevant available data here */
	unsigned short edgeTimeStamp = TC0;			/* Save the timestamp */
	unsigned char PTITCurrentState = PTIT;		/* Save the values on port T regardless of the state of DDRT */
//	unsigned char PORTBCurrentState = PORTB;	/* Save ignition output state for first bank */
//	unsigned char PORTACurrentState = PORTA;	/* Save ignition output state for second bank */
	
	/* Set up edges as per config */
	unsigned char risingEdge;
	if(coreSettingsA & PRIMARY_POLARITY){
		risingEdge = PTITCurrentState & 0x01; 
	}else{
		risingEdge = !(PTITCurrentState & 0x01); 
	}
	
	if(risingEdge){
		// echo input condition to output LED adjacent on JimStim
		PORTJ |= 0x80;

		// increment crank pulses
		primaryPulsesPerSecondaryPulse++;
		
		// don't run until the second trigger has come in and the period is correct (VERY temporary)
		if(!(coreStatusA & PRIMARY_SYNC) || !(coreStatusA & RPM_VALID)){
			return;
		}
		
		// find out what our timer value means
		unsigned long currentExt;
		if(TFLGOF && !(edgeTimeStamp & 0x8000)){ /* see 10.3.5 paragraph 4 of 68hc11 ref manual for details */
			currentExt = timerExtension + 1; 
		}else{
			currentExt = timerExtension;
		}

		// make our time value 32 bit
		unsigned long currentTime = currentExt << 16; // leave the low word all zeros and move the timer extension to the high word
		currentTime += edgeTimeStamp; // replace the zeros in the low word with the contents of the timer capture

		// get the data we actually want
		//engineCyclePeriod = currentTime - lastSecondaryOddTimeStamp; // save the period			
		//lastSecondaryOddTimeStamp = currentTime; // save this stamp for next time round
		
		// temporary data from inputs
		primaryLeadingEdgeTimeStamp = edgeTimeStamp;
		timeBetweenSuccessivePrimaryPulses = lastPrimaryPulseTimeStamp - primaryLeadingEdgeTimeStamp;
		lastPrimaryPulseTimeStamp = primaryLeadingEdgeTimeStamp;
		timeBetweenSuccessivePrimaryPulsesBuffer = (timeBetweenSuccessivePrimaryPulses >> 1) + (timeBetweenSuccessivePrimaryPulsesBuffer >> 1);

		// TODO make scheduling either fixed from boot with a limited range, OR preferrably if its practical scheduled on the fly to allow arbitrary advance and retard of both fuel and ignition.

		/* Check for loss of sync by too high a count */
		if(primaryPulsesPerSecondaryPulse > 12){
			/* Increment the lost sync count */
			lostCrankSyncCounter++;
			
			/* Clear synced status */
			coreStatusA &= ~PRIMARY_SYNC;
			
			/* Clear the rpm valid flag */
			coreStatusA &= ~RPM_VALID;
			
			/* Reset the count of teeth */
			primaryPulsesPerSecondaryPulse = 0;
			
			/* Get the hell out of here before we do something bad */
			return;
		}
		
		// CAUTION came to me lying in bed half asleep idea :
		
		// TODO move tooth selection to the calc loop in main such that this routine just iterates through an array of events and schedules those that are destined for this tooth. 

		// if ign enabled
			// iterate through ignition first, schedule all of those
			// iterate through dwell next, schedule all of those
		// if fuel enabled
			// iterate through main fuel next, schedule all of those
			// if staging enabled and required
				// iterate through staged fuel last,
		
		// TODO should make for a clean compact scheduling implementation. the fuel code doesn't care when/how it has started in the past, and hopefully ign will be the same.
		
		
//		short modulusGenerator = 12/ports; TODO make it cyl count generic BUT NOT HERE
//		if((primaryPulsesPerSecondaryPulse % modulusGenerator) == 0){
		if(((primaryPulsesPerSecondaryPulse % 2) == 0) && (totalPulseWidth > injectorMinimumPulseWidth)){ // this needs to be per channel. TODO

			/* Determine if half the cycle is bigger than short-max */
			unsigned short maxAngleAfter;
			if((engineCyclePeriod >> 1) > 0xFFFF){
				maxAngleAfter = 0xFFFF;
			}else{
				maxAngleAfter = (short)(engineCyclePeriod >> 1);
			}
			
			/* Check advance to ensure it is less than 1/2 of the previous engine cycle and more than codetime away */
			unsigned short advance;
			if(totalAngleAfterReferenceInjection > maxAngleAfter){ // if too big, make it max
				advance = maxAngleAfter;
			}else if(totalAngleAfterReferenceInjection < trailingEdgeSecondaryRPMInputCodeTime){ // if too small, make it min
				advance = trailingEdgeSecondaryRPMInputCodeTime;
			}else{ // else use it as is
				advance = totalAngleAfterReferenceInjection;
			}
			
			// determine the long and short start times
			unsigned short startTime = primaryLeadingEdgeTimeStamp + advance;
			unsigned long startTimeLong = currentTime + advance;

			/* Determine the channels to schedule */
			unsigned char fuelChannel = (primaryPulsesPerSecondaryPulse / 2) - 1;
			unsigned char ignitionChannel = (primaryPulsesPerSecondaryPulse / 2) - 1;
			
			// determine whether or not to reschedule
			unsigned char reschedule = 0;
			unsigned long diff = startTimeLong - (injectorMainEndTimes[fuelChannel] + injectorSwitchOffCodeTime);
			if(diff > LONGHALF){
				reschedule = 1; // http://www.diyefi.org/forum/viewtopic.php?f=8&t=57&p=861#p861
			}

			// schedule the appropriate channel
			if(!(*injectorMainControlRegisters[fuelChannel] & injectorMainEnableMasks[fuelChannel]) || reschedule){ /* If the timer isn't still running, or if its set too long, set it to start again at the right time soon */
				*injectorMainControlRegisters[fuelChannel] |= injectorMainEnableMasks[fuelChannel];
				*injectorMainTimeRegisters[fuelChannel] = startTime;
				TIE |= injectorMainOnMasks[fuelChannel];
				TFLG = injectorMainOnMasks[fuelChannel];
			}else{
				injectorMainStartTimesHolding[fuelChannel] = startTime;
				selfSetTimer |= injectorMainOnMasks[fuelChannel]; // setup a bit to let the timer interrupt know to set its own new start from a var
			}

			// TODO advance/retard/dwell numbers all need range checking etc done. some of this should be done in the calculator section, and some here. currently none is done at all and for that reason, this will not work in a real system yet, if it works at all.
			// as do array indexs here and in the ISRs...
			
			
			// TODO implement mechanism for dropping a cylinder in event of over queueing or spark cut/round robin
			// important as ignition sequence disrupted when this occurs as it stands.
			
			// TODO check queue length checks to ensure we dont count up to somewhere we can never count down from. This could be causing the hanging long phenomina
			
			// DWELL
			
			// If dwell is not currently enabled, set it all up
			if(!(PITCE & DWELL_ENABLE)){
				/* Schedule Dwell event (do this first because it comes earliest. */
				// set the channel to fire
				nextDwellChannel = ignitionChannel;

				// set the time
				PITLD0 = advance;
//				PITLD0 = ignitionAdvances[ignitionChannel + outputBankIgnitionOffset] - currentDwell[outputBank]; BAD for various reasons!

				// clear the flags first as they apparently become set any old time whether enabled or not.
				PITTF |= DWELL_ENABLE;

				// turn on the ints
				PITINTE |= DWELL_ENABLE;

				// clear the flags first as they apparently become set any old time whether enabled or not.
				PITTF |= DWELL_ENABLE;

				// enable channels
				PITCE |= DWELL_ENABLE;
			}else if(dwellQueueLength == 0){
				// load time offset such that next period is correct
				PITLD0 = (advance - PITCNT0);
				
				// increment queue length
				dwellQueueLength++;
			}else if(dwellQueueLength > combustionEventsPerEngineCycle){ //TODO sensible figures here for array index OOBE
				// do nothing, or increment a counter or something similar.
			}else{
				unsigned short sumOfDwells = PITLD0;
				// add up the prequeued time periods
				
				// queue = 1 pitld is all
				// queue = 2 one from 0 index of array AND pitld
				
				unsigned char index = 0;
				while(index < (dwellQueueLength -1)){
					sumOfDwells += queuedDwellOffsets[index];
					index++;
				}
//				for(index = 0;index < (dwellQueueLength -1);index++){ // is this right?
//					sumOfDwells += queuedDwellOffsets[index];
//				}
				
				// store time offset in appropriate array location
				queuedDwellOffsets[dwellQueueLength - 1] = advance - (PITCNT0 + sumOfDwells);
				
				// increment queue length from one or more
				dwellQueueLength++;
			}

			// IGNITION experimental stuff
			
			// If ignition is not currently enabled, set it all up
			if(!(PITCE & IGNITION_ENABLE)){
				/* Schedule Ignition event (do this first because it comes earliest. */
				// set the channel to fire
				nextIgnitionChannel = ignitionChannel;

				// figure out the time to set the delay reg to
				PITLD1 = advance + injectorMainPulseWidths[fuelChannel];
//				PITLD1 = ignitionAdvances[ignitionChannel + outputBankIgnitionOffset];

				// clear the flags first as they apparently become set any old time whether enabled or not.
				PITTF |= IGNITION_ENABLE;

				// turn on the ints
				PITINTE |= IGNITION_ENABLE;

				// clear the flags first as they apparently become set any old time whether enabled or not.
				PITTF |= IGNITION_ENABLE;

				// enable channels
				PITCE |= IGNITION_ENABLE;
			}else if(ignitionQueueLength == 0){
				// load timer register
				PITLD1 = ((advance + injectorMainPulseWidths[fuelChannel]) - PITCNT1);

				// increment to 1 
				ignitionQueueLength++;
			}else if(ignitionQueueLength > combustionEventsPerEngineCycle){ //TODO sensible figures here for array index OOBE
				// do nothing, or increment a counter or something similar.
			}else{		
				unsigned short sumOfIgnitions = PITLD1;
				// add up the prequeued time periods
				
				// queue = 1 pitld is all
				// queue = 2 one from 0 index of array AND pitld
				
				
				unsigned char index = 0;
				while(index < (ignitionQueueLength - 1)){
					sumOfIgnitions += queuedIgnitionOffsets[index];
					index++;
				}
//				for(index = 0;index < (ignitionQueueLength -1);index++){ // is this right?
//					sumOfIgnitions += queuedIgnitionOffsets[index];
//				}
				
				// store time offset in appropriate array location
				queuedIgnitionOffsets[ignitionQueueLength - 1] = advance - (PITCNT1 + sumOfIgnitions);
								
				// increment from 1 or more
				ignitionQueueLength++;
			}
	
			
			// TODO sample ADCs on teeth other than that used by the scheduler in order to minimise peak run time and get clean signals

			/* Read ADCs into the correct bank */
			iat[inputBank] = ATD0DR0;
			clt[inputBank] = ATD0DR1;
			tps[inputBank] = ATD0DR2;
			ego[inputBank] = ATD0DR3;
			map[inputBank] = ATD0DR4;
			spareADC5[inputBank] = ATD0DR5;
			spareADC6[inputBank] = ATD0DR6;
			spareADC7[inputBank] = ATD0DR7;

			// TODO average them here or maybe do that in main as well, and just miss them out if too slow?

//			newVal var word        ' the value from the ADC
//			smoothed var word    ' a nicely smoothed result 
//
//			if newval > smoothed then
//			        smoothed = smoothed + (newval - smoothed)/alpha
//			else
//			        smoothed = smoothed - (smoothed - newval)/alpha
//			endif
	
			// from : http://www.tigoe.net/pcomp/code/category/code/arduinowiring/41			

			/* Set flag to say calc required */
			coreStatusA |= CALC_FUEL_IGN;
			
			/* Reset the counter for reading timeout */
			readingCounter = 0;
		}
	}else{
		PORTJ &= 0x7F;
	}
	
	primaryTeethCounter++;
	// suss out rpm and accurate TDC reference
	
	// if you say it quick, it doesn't sound like much :
	// schedule fuel and ign based on spark cut and fuel cut and timing vars and status vars config vars

	/* If the flag is not cleared at the beginning then the interrupt gets rescheduled while it is running, hence it can't be done at the end of the ISR */
}

void SecondaryRPMISR(void)
{
	/* Clear the interrupt flag for this input compare channel */
	TFLG = 0x02;

	/* Save all relevant available data here */
	unsigned short edgeTimeStamp = TC1;			/* Save the timestamp */
	unsigned char PTITCurrentState = PTIT;		/* Save the values on port T regardless of the state of DDRT */
//	unsigned char PORTBCurrentState = PORTB;	/* Save ignition output state for first bank */
//	unsigned char PORTACurrentState = PORTA;	/* Save ignition output state for second bank */
	
	// TODO discard narrow ones!
	
	/* Set up edges as per config */
	unsigned char risingEdge;
	if(coreSettingsA & SECONDARY_POLARITY){
		risingEdge = PTITCurrentState & 0x02; 
	}else{
		risingEdge = !(PTITCurrentState & 0x02); 
	}
	
	if(risingEdge){
		// echo input condition to output LED adjacent on JimStim
		PORTJ |= 0x40;

		// display the crank pulses
		PORTM = (char)primaryPulsesPerSecondaryPulseBuffer;

		primaryPulsesPerSecondaryPulseBuffer = primaryPulsesPerSecondaryPulse;
		primaryPulsesPerSecondaryPulse = 0;
		
		// if we didn't get the right number of pulses drop sync and start over
		if((primaryPulsesPerSecondaryPulseBuffer != 12) && (coreStatusA & PRIMARY_SYNC)){
			coreStatusA &= ~PRIMARY_SYNC;
			coreStatusA &= ~RPM_VALID;
			lostCrankSyncCounter++;
		}

		// find out what our timer value means
		unsigned long currentExt;
		if(TFLGOF && !(edgeTimeStamp & 0x8000)){ /* see 10.3.5 paragraph 4 of 68hc11 ref manual for details */
			currentExt = timerExtension + 1; 
		}else{
			currentExt = timerExtension;
		}

		// make our time value 32 bit
		unsigned long currentTime = currentExt << 16; // leave the low word all zeros and move the timer extension to the high word
		currentTime += edgeTimeStamp; // replace the zeros in the low word with the contents of the timer capture

		// get the data we actually want
		engineCyclePeriod = currentTime - lastSecondaryOddTimeStamp; // save the period			
		lastSecondaryOddTimeStamp = currentTime; // save this stamp for next time round

		// calculate rough rpm (this will be wrong when the var is used correctly)
		rpm[inputBank] = 75000000 / engineCyclePeriod; // 0.8us ticks, 75mil = 60 seconds

		/* If synced last time, this time make rpm valid */
		if(coreStatusA & PRIMARY_SYNC){
			coreStatusA |= RPM_VALID;
		}
		
		// Because this is our only reference, each time we get this pulse, we know where we are at (simple mode so far)
		coreStatusA |= PRIMARY_SYNC;
	}else{
		PORTJ &= 0xBF;
	}

	secondaryTeethCounter++;
	// suss out phase/engine cycle reference showing which bank we are on

	/* If the flag is not cleared at the beginning then the interrupt gets rescheduled while it is running, hence it can't be done at the end of the ISR */
}
