/*	init.c

	Copyright 2008 Fred Cooke

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

#include "init.h"

/* Set the Phase Locked Loop to our desired frequency (80MHz) and switch to using it for clock (40MHz bus speed) */ 
void init_pll(void)
{
	CLKSEL &= PLLSELOFF;	/* Switches to base external OSCCLK to ensure PLL is not being used (off out of reset, but not sure if the monitor turns it on before passing control or not) */
	PLLCTL &= PLLOFF;		/* Turn the PLL device off to adjust its speed (on by default out of reset) */
	REFDV = PLLDIVISOR;		/* 16MHz / (3 + 1) = 4MHz Bus frequency */
	SYNR = PLLMULTIPLIER;	/* 4MHz * (9 + 1) = 40MHz Bus frequency */
	PLLCTL |= PLLON;		/* Turn the PLL device back on again at 80MHz */

	while (!(CRGFLG & PLLLOCK)){
		/* Do nothing while we wait till the PLL loop locks onto the target frequency. */
		/* Target frequency is given by (2 * (crystal frequency / (REFDV + 1)) * (SYNR + 1)) */
		/* Bus frequency is half PLL frequency and given by ((crystal frequency / (REFDV + 1)) * (SYNR + 1)) */
	}

	CLKSEL = PLLSELON;		/* Switches to PLL clock for internal bus frequency */
	/* from MC9S12XDP512V2.pdf Section 2.4.1.1.2 page 101 Third paragraph */
	/* This takes a MAXIMUM of 4 OSCCLK clock cylces PLUS 4 PLL clock cycles */
	/* During this time ALL clocks freeze, and CPU activity ceases */
	/* For this reason there is no point in waiting for this to occur, we already are */
}

/* TODO Check the code for flash errors based on stored checksum */
void check_checksum(void)
{
	// 
}

/* Configure all the I/O to default values to keep power use down etc */
void init_io(void)
{
	/* for now, hard code all stuff to be outputs as per Freescale documentation,	*/
	/* later what to do will be pulled from flash configuration such that all		*/
	/* things are setup at once, and not messed with thereafter. when the port		*/
	/* something uses is changed via the tuning interface, the confuration will be	*/
	/* done on the fly, and the value burned to flash such that next boot happens	*/
	/* correctly and current running devices are used in that way.					*/
	
	/* Turn off and on and configure all the modules in an explicit way */
	// TODO set up and turn off all modules (CAN,SCI,SPI,IIC,etc)
	
	/* Turn off the digital input buffers on the ATD channels */
	ATD0DIEN = ZEROS; /* You are out of your mind if you waste this on digital Inputs */
	ATD1DIEN0 = ZEROS; /* You are out of your mind if you waste this on digital Inputs (NOT-bonded, can't use) */	
	ATD1DIEN1 = ZEROS; /* You are out of your mind if you waste this on digital Inputs */
	/* TODO Second half of ATD1 - can we disable this somehow */
	
	/* And configure them all for analog input */
	ATD0CTL2 = 0x80; /* Turns on the ADC block. */
	ATD0CTL3 = 0x40; /* Set sequence length = 8 */
	ATD0CTL5 = 0xB0; /* Sets justification to right, multiplex and scan all channels. */
	// TODO find out if this is the default (i suspect it is)
	// TODO look into sampling techniques
	
	/* Set up the PWM component and initialise its values to off */
	PWME = 0x7F; /* Turn on PWM 0 - 6 (7 is user LED on main board) */
	PWMCLK = ZEROS; /* The fastest we can go for all channels */
	PWMPRCLK = ZEROS; /* The fastest prescaler we can go for all channels */
	PWMSCLA = ZEROS; /* The fastest we can go */
	PWMSCLB = ZEROS; /* The fastest we can go */
	/* TODO PWM channel concatenation for high resolution */
	// join channel pairs together here (needs 16 bit regs enabled too)
	/* TODO Initialise pwm channels with frequency, and initial duty for real use */
	// intitial PWM settings for testing
	/* PWM periods */
	PWMPER0 = 0xFF; // 255 for ADC0 testing
	PWMPER1 = 0xFF; // 255 for ADC1 testing
	PWMPER2 = 0xFF; // 255 for ADC1 testing
	PWMPER3 = 0xFF; // 255 for ADC1 testing
	PWMPER4 = 0xFF; // 255 for ADC1 testing
	PWMPER5 = 0xFF; // 255 for ADC1 testing
	PWMPER6 = 0xFF; // 255 for ADC1 testing
	PWMPER7 = 0xFF; // 255 for ADC1 testing
	/* PWM duties */
	PWMDTY0 = 0;
	PWMDTY1 = 0;
	PWMDTY2 = 0;
	PWMDTY3 = 0;
	PWMDTY4 = 0;
	PWMDTY5 = 0;
	PWMDTY6 = 0;
	PWMDTY7 = 0;

	
	/* TODO Set up the SCI0 interface for basic comms here */

	
	/* Initialise the Data Direction Registers */
	/* To outputs based on the note at the end of chapter 1.2.2 of MC9S12XDP512V2.pdf */
	DDRA = ONES; /* GPIO (8) */
	DDRB = ONES; /* GPIO (8) */
	DDRE = 0xFC; /* 0b_1111_1100 : Clock and mode pins PE0,PE1 are input only pins, the rest are GPIO */
	DDRK = ONES; /* Only 0,1,2,3,4,5,7, NOT 6 (7) */
	DDRS = ONES; /* SCI0, SCI1, SPI0 (8) */
	DDRT = 0xFC; /* 0b_1111_1100 set ECT pins 0,1 to IC and 2:7 to OC (8) */
	DDRM = ONES; /* CAN 0 - 3 (8) */
	DDRP = ONES; /* PWM pins (8) */
	DDRH = ZEROS; /* All pins configured as input for misc isrs (SPI1, SPI2) (8) */
	DDRJ = ONES; /* Only 0,1,6,7 are brought out on the 112 pin chip (4) */
	/* Configure the non bonded pins to output to avoid current drain (112 pin package) */ 
	DDRC = ONES; /* NON-bonded external data bus pins */
	DDRD = ONES; /* NON-bonded external data bus pins */
	/* AD0DDR1 You are out of your mind if you waste this on digital Inputs */
	/* AD1DDR1 You are out of your mind if you waste this on digital Inputs */

	/* Initialise the state of pins configured as output */
	/* Initialise to low such that transistor grounded things are all turned off by default. */
	PORTA = ZEROS; /* The serial monitor pin is on 0x40, and could cause problems if capacitance at the output is large when a reset occurs. */
	PORTB = ZEROS; /* Init the rest of the spark outputs as off */
	PORTE = 0x1F; /* 0b_0001_1111 : when not in use 0b_1001_1111 PE7 should be high PE5 and PE6 should be low, the rest high */
	PORTK = ZEROS;
	PORTS = ZEROS;
	PORTT = ZEROS; /* All pins in off state at boot up (only matters for 2 - 7) */ 
	PORTM = ZEROS;
	PORTP = ZEROS; // TODO hook these up to the adc channels such that you can vary the brightness of an led with a pot.
	PORTH = ZEROS;
	PORTJ = ZEROS;
	/* AD0PT1 You are out of your mind if you waste this on digital Inputs */
	/* AD1PT1 You are out of your mind if you waste this on digital Inputs */
}

/* TODO Load the pages of tuning data etc from flash to paged ram for fast access on the fly */
void init_paged_ram(void)
{
	// 
}

void init_stats(void){
	/* Zero all the counters for various conditions */
	UISRCounter = 0;
	lowVoltISRCounter = 0;
	readingCounter = 0;
	lostCamSyncCounter = 0;
	lostCrankSyncCounter = 0;
	lostRPMValidityCounter = 0;
	primaryTeethCounter = 0;
	secondaryTeethCounter = 0;

	/* Real Time Clocks */
	millisToTenths = 0;
	tenthsToSeconds = 0;
	secondsToMinutes = 0;
	realTimeCounterMain = 0;
	realTimeCounterMillis = 0;
	realTimeCounterTenths = 0;
	realTimeCounterSeconds = 0;
	realTimeCounterMinutes = 0;
}

/* Initialise and set up some running variables etc here */
void init_variables(void)
{	
	/* State */
	coreStatusA = ZEROS16;			/* See definitions in freeems.h */
	mainOn = ZEROS;
	stagedOn = ZEROS;
	dwellOn = ZEROS16;
	selfSetTimer = ZEROS;
	rescheduleFuelFlags = ZEROS;
	
	// Values for testing
	totalPulseWidth = 10;
	totalAngleAfterReferenceIgnition = 540;
	totalAngleAfterReferenceInjection = 180;

//	dwellLength = 32000;
//	dwellPeriod = 64000;
	
	/* Set the banks to be the same at boot */
	inputBank = 0;
	outputBank = 0;
	outputBankInjectionOffset = 0;
	outputBankIgnitionOffset = 0;

	/* Setup the pointers to the registers for fueling use, this does NOT work if done in global.c, I still don't know why. */
	injectorMainTimeRegisters[0] = TC2_ADDR;
	injectorMainTimeRegisters[1] = TC3_ADDR;
	injectorMainTimeRegisters[2] = TC4_ADDR;
	injectorMainTimeRegisters[3] = TC5_ADDR;
	injectorMainTimeRegisters[4] = TC6_ADDR;
	injectorMainTimeRegisters[5] = TC7_ADDR;
	injectorMainControlRegisters[0] = TCTL2_ADDR;  
	injectorMainControlRegisters[1] = TCTL2_ADDR;  
	injectorMainControlRegisters[2] = TCTL1_ADDR;
	injectorMainControlRegisters[3] = TCTL1_ADDR;  
	injectorMainControlRegisters[4] = TCTL1_ADDR;  
	injectorMainControlRegisters[5] = TCTL1_ADDR;  

	/* Ignition stuff */
	dwellQueueLength = 0;
	ignitionQueueLength = 0;
	
	// TODO perhaps read from the ds1302 once at start up and init the values or different ones with the actual time and date then update them in RTI
		
	// TODO do these require init????
	primaryPulsesPerSecondaryPulse = 0;
	primaryPulsesPerSecondaryPulseBuffer = 0;
	primaryLeadingEdgeTimeStamp = 0;
	timeBetweenSuccessivePrimaryPulses = 0;
	timeBetweenSuccessivePrimaryPulsesBuffer = 0;
	lastPrimaryPulseTimeStamp = 0;
	lastSecondaryOddTimeStamp = 0;
	engineCyclePeriod = 0; /* TODO Start at ??? sensible init value required, this isn't it i don't think */
	/* TODO this is being misused in engine pos isr as time per rev... */
	
	// TODO remove this : Test stuff
	portHDebounce = 0;
}

/* TODO Load and calculate all configuration data required to run */
void init_configuration(void)
{
	coreSettingsA = 0x07F0;			/* See definitions in freeems.h */

	/* Settings */
	primaryTeeth = 24;
	missingTeeth = 0;
	ports = 6;
	coils = 6;
	revolutionsPerEngineCycle = 2;
	combustionEventsPerEngineCycle = 6;
	capacityOfAirPerCombustionEvent = 500;
	perPrimaryInjectorChannelFlowRate = 550;
	perSecondaryInjectorChannelFlowRate = 550;
	
	readingTimeout = 100;
}

/* Set up the timer module and its various interrupts */
void init_ect_timer(void)
{
	
	// TODO rearrange the order of this stuff and pull enable and int enable out to the last function call of init.
	
	/* Timer channel interrupts */
	TIE = 0x03; /* 0,1 IC interrupts enabled for reading engine position and RPM, 6 OC channels disabled such that no injector switching happens till scheduled */
	TFLG = ONES; /* Clear all the flags such that we are up and running before they first occur */
	TFLGOF = ONES; /* Clear all the flags such that we are up and running before they first occur */

	/* TODO Turn the timer on and set the rate and overflow interrupt */
	TSCR1 = 0x88; /* 0b_1000_1000 Timer enabled, and precision timer turned on */
	TSCR2 = 0x87; /* 0b_1000_0111 Overflow interrupt enable, divide by 256 if precision turned off */
//	PTPSR = 0x03; /* 4 prescaler gives .1uS resolution and max period of 7ms measured */
	PTPSR = 0x1F; /* 32 prescaler gives 0.8uS resolution and max period of 52.4288ms measured */
//	PTPSR = 0x3F; /* 64 prescaler gives 1.6uS resolution and max period of 105ms measured */
//	PTPSR = 0xFF; /* 256 prescaler gives 6.4uS resolution and max period of 400ms measured */
//	PTPSR = 0x7F; /* 128 prescaler gives 3.2uS resolution and max period of 200ms measured */
	/* http://www.google.com/search?hl=en&safe=off&q=1+%2F+%2840MHz+%2F+32+%29&btnG=Search */
	/* http://www.google.com/search?hl=en&safe=off&q=1+%2F+%2840MHz+%2F+32+%29+*+2%5E16&btnG=Search */
	/* www.mecheng.adelaide.edu.au/robotics_novell/WWW_Devs/Dragon12/LM4_Timer.pdf */

	/* Initial actions */
	TIOS = 0xFC; /* 0b_1111_1100 0 and 1 are input capture, 2 through 7 are output compare */
	TCTL1 = ZEROS; /* Set disabled at startup time, use these and other flags to switch fueling on and off inside the decoder */
	TCTL2 = ZEROS; /* 0,1 have compare turned off regardless as they are in IC mode. */
	TCTL3 = ZEROS; /* Capture off for 4 - 7 */
	TCTL4 = 0x0F; /* Capture on both edges of two pins for IC (0,1), capture off for 2,3 */
	
	// TODO setup delay counters on 0 and 1 to filter noise (nice feature!)
	//DLYCT = ??; built in noise filter
	
	// TODO modulus timer test stuff
	// to flash the user led...
	PTMCPSR = ONES; // Precision prescaler slowest/longest possible
	MCCNT = ONES16; // already like this out of reset, but, may want to init it shorter here?
	MCCTL = 0xC4; // turn on and setup the mod down counter
	MCFLG = 0x80; // clear the flag up front
	
}

/* Configure the PIT timers for their various uses. */
void init_pit_timer(void){
	/*  */
	// set micro periods
	PITMTLD0 = 0x1F; /* 32 prescaler gives 0.8uS resolution and max period of 52.4288ms measured */
	PITMTLD1 = 0x1F; /* ditto */
	/* http://www.google.com/search?hl=en&safe=off&q=1+%2F+%2840MHz+%2F+32+%29 Exactly the same as for ECT */
	
	// set timers running
//	PITLD0 = dwellPeriod;
	// enable module
	PITCFLMT = 0x80;
	// enable channels
	//PITCE = 0x03;
	// enable interrupt
//	PITINTE = 0x01;
	// clear flags
	//PITFLT = ONES;
}

/* Setup the sci module(s) that we need to use. */
void init_sci_stuff(void)
{
	/* The alternative register set selector defaults to zero */

	// set the baud/data speed
	SCI0BD = baudDivisor115p2;

	// etc

	/* Switch to alternative register set? */

	// etc

	/* Switch back again? */

	/* 
	 * 0 = LOOPS (normal two wire operation)
	 * 0 = SCISWAI (Wait mode on)
	 * 0 = RSRC (if loops=1, int/ext wiring)
	 * 1 = M MODE (9 bit operation)
	 * 0 = WAKE (idle line wakeup)
	 * 0 = ILT (idle line type count start pos)
	 * 1 = PE (parity on)
	 * 1 = PT (odd parity) (minicom defaults to no parity) 
	 * 
	 * 00010011 = 0x12
	 */
	SCI0CR1 = 0x13;

	/*
	 * 1 = TIE (tx data empty isr)
	 * 1 = TCIE (tx complete isr)
	 * 0 = RIE (rx full isr) 
	 * 0 = ILIE (idle line isr)
	 * 1 = TE (transmit enabled)
	 * 0 = RE (receive disabled)
	 * 0 = RWU (rx wake up normal)
	 * 0 = SBK (send break off)
	 * 
	 * 11001000 = 0xC8
	 * 00001000 = 0x08 temporarily for the purpose of sending test data.
	 */ 
	SCI0CR2 = 0x08;
}

/* Set up all the remaining interrupts */
void init_interrupts(void)
{
	/* IMPORTANT : Set the s12x vector base register (Thanks Karsten!!) */
	IVBR = 0xF7; /* Without this, the interrupts will never find your code! */
	
	/* Set up the Real Time Interrupt */
	RTICTL = 0x81; /* 0b_1000_0001 0.125ms/125us period http://www.google.com/search?hl=en&safe=off&q=1+%2F+%2816MHz+%2F+%282+*+10%5E3%29+%29&btnG=Search */
//	RTICTL = 0xF9; /* 0b_1111_1001 0.125s/125ms period http://www.google.com/search?hl=en&safe=off&q=1+%2F+%2816MHz+%2F+%282*10%5E6%29+%29&btnG=Search */
	CRGINT |= 0x80; /* Enable the RTI */
	CRGFLG = 0x80; /* Clear the RTI flag */
	
	// set up port H for testing
	PPSH = ZEROS; // falling edge/pull up for all
	PIEH = ONES; // enable all pins interrupts
	PIFH = ONES; // clear all port H interrupt flags
	
	// TODO set up irq and xirq for testing
	// IRQCR for IRQ
	// 
	
	/* VReg API setup (only for wait mode? i think so) */
//	VREGAPIR = 0x09C3; /* For 500ms period : (500ms - 0.2ms) / 0.2ms = 0b100111000011 = 2499 */
//	VREGAPICL = 0x02; /* Enable the interrupt */
//	VREGAPICL = 0x04; /* Start the counter running */
	/* Writing a one to the flag will set it if it is unset, so best not to mess with it here as it probably starts off unset */
	
	/* LVI Low Voltage Interrupt enable */
	VREGCTRL = 0x02; // Counts bad power events for diagnosis reasons
}

/* Main init function to be called from main.c before enterring the main loop */
void init(void){
	ATOMIC_START();			/* Disable ALL interrupts while we configure the board ready for use */
	init_pll();				/* Set up the PLL and use it */
	check_checksum();		/* TODO maybe checksum the code base excluding the areas of flash used for tuning data OR, include the tuning data and write a checksum in with it each time. display a warning via tacho, shiftlights, etc if stuffed. */
	init_io();				/* TODO make this config dependent. Set up all the pins and modules to be in low power harmless states */
	init_paged_ram();		/* TODO load the blocks of data flash to the paged ram blocks for uber fast data lookup */
	init_stats();			/* Initialise all statistics variables */
	init_variables();		/* Initialise the rest of the running variables etc */
	init_configuration();	/* TODO Set user/feature/config up here! */
	init_ect_timer();		/* TODO move this to inside config in an organised way. Set up the timer module and its various aspects */
	init_pit_timer();		/* TODO ditto... */
	init_sci_stuff();		/* Setup the sci module(s) that we will use. */
	init_interrupts();		/* still last, reset timers, enable interrupts here TODO move this to inside config in an organised way. Set up the rest of the individual interrupts */
	ATOMIC_END(); /* Re-enable any configured interrupts */
}
